/* -------------------------------------------------------------------------
 * Copyright (c) 2009 Steve Jolly. All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See the COPYING file for a copy of the GNU General Public Licence.
 *
 * -------------------------------------------------------------------------
 */
package net.elvum.hpgen;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import javax.microedition.rms.*;

/**
 * HPGenPrefs: a singleton that acts as a gateway to the persistently
 * stored preferences (using the RMS)
 * @author Steve Jolly
 */
public class HPGenPrefs {

    private static HPGenPrefs instance;
    private RecordStore rsDomains;
    private RecordStore rsPrefs;
    public static final byte MD5 = 0;
    public static final byte SHA1 = 1;
    private byte pHashFunction = SHA1; //default
    public static final byte HEX_LOWER = 0;
    public static final byte HEX_UPPER = 1;
    public static final byte BASE64 = 2;
    public static final byte BASE64_NO1A = 3;
    private byte pDisplayMode = BASE64; //default
    private static final byte PREFS_FORMAT_VERSION = 2;
    private byte pLength = 8;
    public static final byte NO = 0;
    public static final byte YES = 1;
    private byte pStoreDomains = YES; //default
    private int prefsRecordID = -1;
    private static final int DOMAINS_FORMAT_VERSION = 0;
    private int positionIncrementor = 0;

    private HPGenPrefs() {
        try {
            rsDomains = RecordStore.openRecordStore("net.elvum.hpgen.domains", true);
            rsPrefs = RecordStore.openRecordStore("net.elvum.npgen.prefs", true);
        } catch (RecordStoreException ex) {
            ex.printStackTrace();
        }
        try {
            readPrefs();
            getDomainList();
        } catch (RecordStoreException ex) {
            ex.printStackTrace();
        //Ignore this exception - we'll just have default preferences.
        }
    }

    /**
     * Singleton accessor method
     * @return the singleton instance of the HPGenPrefs class
     */
    public static synchronized HPGenPrefs getInstance() {
        if (instance == null) {
            instance = new HPGenPrefs();
        }
        return instance;
    }

    /**
     * read in the preferences from persistent storage (RMS)
     */
    private void readPrefs() throws RecordStoreException {
        RecordEnumeration prefs = null;
        if (prefsRecordID == -1) {
            prefs = rsPrefs.enumerateRecords(null, null, false);
            int id = -1;
            while (prefs.hasNextElement()) {
                id = prefs.nextRecordId();
            }
            if (id == -1) {
                return;
            }
            prefsRecordID = id;
        }
        byte[] prefsBytes = rsPrefs.getRecord(prefsRecordID);
        switch (prefsBytes[0]) {
            case 0:
                pHashFunction = prefsBytes[1];
                pDisplayMode = prefsBytes[2];
                break;
            case 1:
                pHashFunction = prefsBytes[1];
                pDisplayMode = prefsBytes[2];
                pLength = prefsBytes[3];
                break;
            case 2:
                pHashFunction = prefsBytes[1];
                pDisplayMode = prefsBytes[2];
                pLength = prefsBytes[3];
                pStoreDomains = prefsBytes[4];
                break;
            default:
                return;
        }
        prefs.destroy();
    }

    /**
     * Write the preferences to persistent storage (RMS)
     */
    private void writePrefs() throws RecordStoreException {
        if (prefsRecordID != -1) {
            rsPrefs.deleteRecord(prefsRecordID);
        }
        byte[] encodedPrefs = new byte[5];
        encodedPrefs[0] = PREFS_FORMAT_VERSION;
        encodedPrefs[1] = pHashFunction;
        encodedPrefs[2] = pDisplayMode;
        encodedPrefs[3] = pLength;
        encodedPrefs[4] = pStoreDomains;
        prefsRecordID = rsPrefs.addRecord(encodedPrefs, 0, encodedPrefs.length);
    }

    /**
     * getter method for the stored hashfunction selection
     * @return a byte corresponding to the type-unsafe enumerated hashfunction
     * currently selected
     */
    public byte getHashFunction() {
        return pHashFunction;
    }

    /**
     * Setter function for the hashfunction selection.  Sets the current selection
     * both in memory and persistent storage (RMS)
     * @param hashFunction the byte corresponding to the type-unsafe enumeration
     * of hash functions
     */
    public void setHashFunction(byte hashFunction) {
        pHashFunction = hashFunction;
        try {
            writePrefs();
        } catch (RecordStoreException ex) {
            ex.printStackTrace();
        //We'll just live with it if we can't write the preferences to
        //persistent storage
        }
    }

    /**
     * getter method for the stored display mode selection
     * @return a byte corresponding to the type-unsafe enumerated display mode
     * currently selected
     */
    public byte getDisplayMode() {
        return pDisplayMode;
    }

    /**
     * Setter function for the display mode selection.  Sets the current selection
     * both in memory and persistent storage (RMS)
     * @param displayMode the byte corresponding to the type-unsafe enumeration
     * of display modes
     */
    public void setDisplayMode(byte displayMode) {
        pDisplayMode = displayMode;
        try {
            writePrefs();
        } catch (RecordStoreException ex) {
            ex.printStackTrace();
        //We'll just live with it if we can't write the preferences to
        //persistent storage
        }
    }

    /**
     * getter method for the stored generated password length
     * @return a byte containing the stored length of the generated password
     */
    public byte getLength() {
        return pLength;
    }

    /**
     * setter method for the stored generated password length
     * @param length the new length of the generated password
     */
    public void setLength(byte length) {
        pLength = length;
        try {
            writePrefs();
        } catch (RecordStoreException ex) {
            ex.printStackTrace();
        //We'll just live with it if we can't write the preferences to
        //persistent storage
        }
    }

    /**
     * getter method for the stored storedomains selection
     * @return true if the current selection is to store domains, false otherwise
     */
    public boolean getStoreDomains() {
        return pStoreDomains == YES;
    }

    /**
     * Setter function for the storedomains selection.  Sets the current selection
     * both in memory and persistent storage (RMS)
     * @param storeDomains whether or not to store domains
     */
    public void setStoreDomains(boolean storeDomains) {
        pStoreDomains = storeDomains ? YES : NO;
        try {
            writePrefs();
        } catch (RecordStoreException ex) {
            ex.printStackTrace();
        //We'll just live with it if we can't write the preferences to
        //persistent storage
        }
    }

    /**
     * Getter method for the domains stored in the RMS
     * @return a String array containing the domains stored in the RMS
     */
    public String[] getDomainList() {
        String[] domainList = null;
        RecordComparator comparator = new RecordComparator() {

            public int compare(byte[] rec1, byte[] rec2) {
                int position1 = 0, position2 = 0;
                try {
                    ByteArrayInputStream bin = new ByteArrayInputStream(rec1);
                    DataInputStream din = new DataInputStream(bin);
                    din.readInt();
                    position1 = din.readInt();

                    bin = new ByteArrayInputStream(rec2);
                    din = new DataInputStream(bin);
                    din.readInt();
                    position2 = din.readInt();
                } catch (IOException ex) {
                    ex.printStackTrace();
                }
                if (position1 == position2) {
                    return EQUIVALENT;
                }
                return (position1 < position2) ? PRECEDES : FOLLOWS;

            }
        };

        try {
            RecordEnumeration records = rsDomains.enumerateRecords(null, comparator, false);
            domainList = new String[records.numRecords()];
            int dlIndex = 0;
            while (records.hasNextElement()) {
                int id = records.nextRecordId();
                ByteArrayInputStream bin = new ByteArrayInputStream(rsDomains.getRecord(id));
                DataInputStream din = new DataInputStream(bin);
                try {
                    switch (din.readInt()) { //first int is version of record
                        case 0:
                            int positionIndex = din.readInt();
                            if (positionIndex >= positionIncrementor) {
                                positionIncrementor = positionIndex + 1;
                            }
                            domainList[dlIndex++] = din.readUTF();
                            break;
                    }
                } catch (IOException ex) {
                    ex.printStackTrace(); //if an IO exception occurs, just ignore the record, it's probably an unrecognised format
                }
            }
            records.destroy();
        } catch (RecordStoreException ex) {
            ex.printStackTrace();
        }
        return domainList;
    }

    /**
     * Determine whether or not a given domain name is stored in the RMS
     * @param domainName the domain to test for
     * @return true if the domain is already stored in the RMS; false otherwise
     * @throws RecordStoreException if the RMS cannot be accessed
     */
    public boolean containsDomain(final String domainName) throws RecordStoreException {
        if (domainName == null) {
            return false;
        }
        RecordFilter domainNameMatcher = new RecordFilter() {

            public boolean matches(byte[] candidate) {
                ByteArrayInputStream bin = new ByteArrayInputStream(candidate);
                DataInputStream din = new DataInputStream(bin);
                try {
                    switch (din.readInt()) {
                        case 0:
                            din.readInt(); //skip position
                            return domainName.equals(din.readUTF());
                        default:
                            return false;
                    }
                } catch (IOException ex) {
                    ex.printStackTrace();
                    return false;
                }
            }
        };
        RecordEnumeration matchingRecords = rsDomains.enumerateRecords(domainNameMatcher, null, false);
        return (matchingRecords.numRecords() > 0);
    }

    /**
     * Add the given domain to the list stored in the RMS
     * @param newDomain the new domain name to be stored
     * @throws RecordStoreException if the RMS could not be accessed
     */
    public void addDomain(String newDomain) throws RecordStoreException {
        if (!this.containsDomain(newDomain)) {
            try {
                ByteArrayOutputStream bout = new ByteArrayOutputStream();
                DataOutputStream dout = new DataOutputStream(bout);
                dout.writeInt(DOMAINS_FORMAT_VERSION);
                dout.writeInt(positionIncrementor++);
                dout.writeUTF(newDomain);
                byte[] recordBytes = bout.toByteArray();
                rsDomains.addRecord(recordBytes, 0, recordBytes.length);
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
    }

    /**
     * Deletes a given domain from the list stored in the RMS, if it exists
     * @param domainName the name of the domain to be deleted
     * @throws RecordStoreException if the RMS could not be accessed
     */
    public void deleteDomain(final String domainName) throws RecordStoreException {
        RecordFilter domainNameMatcher = new RecordFilter() {

            public boolean matches(byte[] candidate) {
                ByteArrayInputStream bin = new ByteArrayInputStream(candidate);
                DataInputStream din = new DataInputStream(bin);
                try {
                    switch (din.readInt()) {
                        case 0:
                            din.readInt(); //skip position
                            return domainName.equals(din.readUTF());
                        default:
                            return false;
                    }
                } catch (IOException ex) {
                    ex.printStackTrace();
                    return false;
                }
            }
        };
        RecordEnumeration matchingRecords = rsDomains.enumerateRecords(domainNameMatcher, null, false);
        while (matchingRecords.hasNextElement()) {
            rsDomains.deleteRecord(matchingRecords.nextRecordId());
        }

    }

    /** Add a domain to the list stored in the RMS in such a way that duplicates
     * are precluded.  Note that a side-effect of the implementation is that
     * domain names are ordered in the RMS with the most recently-added via this
     * method being last in the list.
     * @param domainString the domain to be added to the RMS
     * @throws RecordStoreException if the domain could not be added due to
     * the inaccessability of the RMS
     */
    public void exclusiveAddDomain(String domainString) throws RecordStoreException {
        try {
            //for now, deleting and adding again forces the domain to the bottom of the list in the record store
            deleteDomain(domainString);
        } catch (RecordStoreException ex) {
            ex.printStackTrace();
        }

        addDomain(domainString);
    }
}
